package xyz.lychee.lagfixer.libs.adventure.platform.facet;

import java.io.Closeable;
import java.time.Duration;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.function.Supplier;
import xyz.lychee.lagfixer.libs.adventure.audience.MessageType;
import xyz.lychee.lagfixer.libs.adventure.bossbar.BossBar.Color;
import xyz.lychee.lagfixer.libs.adventure.bossbar.BossBar.Flag;
import xyz.lychee.lagfixer.libs.adventure.bossbar.BossBar.Overlay;
import xyz.lychee.lagfixer.libs.adventure.identity.Identity;
import xyz.lychee.lagfixer.libs.adventure.sound.SoundStop;
import xyz.lychee.lagfixer.libs.adventure.text.Component;
import xyz.lychee.lagfixer.libs.annotations.NotNull;
import xyz.lychee.lagfixer.libs.annotations.Nullable;
import xyz.lychee.lagfixer.libs.annotations.ApiStatus.Internal;

@Internal
public interface Facet<V> {
   @SafeVarargs
   static <V, F extends Facet<? extends V>> @NotNull Collection<F> of(final Supplier<F>... suppliers) {
      List<F> facets = new LinkedList();
      Supplier[] var2 = suppliers;
      int var3 = suppliers.length;

      for(int var4 = 0; var4 < var3; ++var4) {
         Supplier<F> supplier = var2[var4];

         Facet facet;
         try {
            facet = (Facet)supplier.get();
         } catch (NoClassDefFoundError var8) {
            Knob.logMessage("Skipped facet: %s", supplier.getClass().getName());
            continue;
         } catch (Throwable var9) {
            Throwable error = var9;
            Knob.logError(error, "Failed facet: %s", supplier);
            continue;
         }

         if (!facet.isSupported()) {
            Knob.logMessage("Skipped facet: %s", facet);
         } else {
            facets.add(facet);
            Knob.logMessage("Added facet: %s", facet);
         }
      }

      return facets;
   }

   static <V, F extends Facet<V>> @Nullable F of(final @Nullable Collection<F> facets, final @Nullable V viewer) {
      if (facets != null && viewer != null) {
         Iterator var2 = facets.iterator();

         while(var2.hasNext()) {
            F facet = (Facet)var2.next();

            try {
               if (facet.isApplicable(viewer)) {
                  Knob.logMessage("Selected facet: %s for %s", facet, viewer);
                  return facet;
               }

               if (Knob.DEBUG) {
                  Knob.logMessage("Not selecting %s for %s", facet, viewer);
               }
            } catch (ClassCastException var5) {
               ClassCastException error = var5;
               if (Knob.DEBUG) {
                  Knob.logMessage("Exception while getting facet %s for %s: %s", facet, viewer, error.getMessage());
               }
            }
         }

         return null;
      } else {
         return null;
      }
   }

   default boolean isSupported() {
      return true;
   }

   default boolean isApplicable(final @NotNull V viewer) {
      return true;
   }

   public interface Pointers<V> extends Facet<V> {
      void contributePointers(final V viewer, final xyz.lychee.lagfixer.libs.adventure.pointer.Pointers.Builder builder);
   }

   public interface TabList<V, M> extends Message<V, M> {
      void send(final V viewer, final @Nullable M header, final @Nullable M footer);
   }

   public interface FakeEntity<V, P> extends Position<V, P>, Closeable {
      void teleport(final @NotNull V viewer, final @Nullable P position);

      void metadata(final int position, final @NotNull Object data);

      void invisible(final boolean invisible);

      void health(final float health);

      void name(final @NotNull Component name);

      void close();
   }

   public interface BossBarEntity<V, P> extends BossBar<V>, FakeEntity<V, P> {
      int OFFSET_PITCH = 30;
      int OFFSET_YAW = 0;
      int OFFSET_MAGNITUDE = 40;
      int INVULNERABLE_KEY = 20;
      int INVULNERABLE_TICKS = 890;

      default void bossBarProgressChanged(final xyz.lychee.lagfixer.libs.adventure.bossbar.@NotNull BossBar bar, final float oldProgress, final float newProgress) {
         this.health(newProgress);
      }

      default void bossBarNameChanged(final xyz.lychee.lagfixer.libs.adventure.bossbar.@NotNull BossBar bar, final @NotNull Component oldName, final @NotNull Component newName) {
         this.name(newName);
      }

      default void addViewer(final @NotNull V viewer) {
         this.teleport(viewer, this.createPosition(viewer));
      }

      default void removeViewer(final @NotNull V viewer) {
         this.teleport(viewer, (Object)null);
      }
   }

   public interface BossBarPacket<V> extends BossBar<V> {
      int ACTION_ADD = 0;
      int ACTION_REMOVE = 1;
      int ACTION_HEALTH = 2;
      int ACTION_TITLE = 3;
      int ACTION_STYLE = 4;
      int ACTION_FLAG = 5;

      default int createColor(final xyz.lychee.lagfixer.libs.adventure.bossbar.BossBar.@NotNull Color color) {
         if (color == Color.PURPLE) {
            return 5;
         } else if (color == Color.PINK) {
            return 0;
         } else if (color == Color.BLUE) {
            return 1;
         } else if (color == Color.RED) {
            return 2;
         } else if (color == Color.GREEN) {
            return 3;
         } else if (color == Color.YELLOW) {
            return 4;
         } else if (color == Color.WHITE) {
            return 6;
         } else {
            Knob.logUnsupported(this, color);
            return 5;
         }
      }

      default int createOverlay(final xyz.lychee.lagfixer.libs.adventure.bossbar.BossBar.@NotNull Overlay overlay) {
         if (overlay == Overlay.PROGRESS) {
            return 0;
         } else if (overlay == Overlay.NOTCHED_6) {
            return 1;
         } else if (overlay == Overlay.NOTCHED_10) {
            return 2;
         } else if (overlay == Overlay.NOTCHED_12) {
            return 3;
         } else if (overlay == Overlay.NOTCHED_20) {
            return 4;
         } else {
            Knob.logUnsupported(this, overlay);
            return 0;
         }
      }

      default byte createFlag(final byte flagBit, @NotNull final Set<xyz.lychee.lagfixer.libs.adventure.bossbar.BossBar.Flag> flagsAdded, @NotNull final Set<xyz.lychee.lagfixer.libs.adventure.bossbar.BossBar.Flag> flagsRemoved) {
         byte bit = flagBit;
         Iterator var5 = flagsAdded.iterator();

         xyz.lychee.lagfixer.libs.adventure.bossbar.BossBar.Flag flag;
         while(var5.hasNext()) {
            flag = (xyz.lychee.lagfixer.libs.adventure.bossbar.BossBar.Flag)var5.next();
            if (flag == Flag.DARKEN_SCREEN) {
               bit = (byte)(bit | 1);
            } else if (flag == Flag.PLAY_BOSS_MUSIC) {
               bit = (byte)(bit | 2);
            } else if (flag == Flag.CREATE_WORLD_FOG) {
               bit = (byte)(bit | 4);
            } else {
               Knob.logUnsupported(this, flag);
            }
         }

         var5 = flagsRemoved.iterator();

         while(var5.hasNext()) {
            flag = (xyz.lychee.lagfixer.libs.adventure.bossbar.BossBar.Flag)var5.next();
            if (flag == Flag.DARKEN_SCREEN) {
               bit &= -2;
            } else if (flag == Flag.PLAY_BOSS_MUSIC) {
               bit &= -3;
            } else if (flag == Flag.CREATE_WORLD_FOG) {
               bit &= -5;
            } else {
               Knob.logUnsupported(this, flag);
            }
         }

         return bit;
      }
   }

   public interface BossBar<V> extends xyz.lychee.lagfixer.libs.adventure.bossbar.BossBar.Listener, Closeable {
      int PROTOCOL_BOSS_BAR = 356;

      default void bossBarInitialized(final xyz.lychee.lagfixer.libs.adventure.bossbar.@NotNull BossBar bar) {
         this.bossBarNameChanged(bar, bar.name(), bar.name());
         this.bossBarColorChanged(bar, bar.color(), bar.color());
         this.bossBarProgressChanged(bar, bar.progress(), bar.progress());
         this.bossBarFlagsChanged(bar, bar.flags(), Collections.emptySet());
         this.bossBarOverlayChanged(bar, bar.overlay(), bar.overlay());
      }

      void addViewer(final @NotNull V viewer);

      void removeViewer(final @NotNull V viewer);

      boolean isEmpty();

      void close();

      @FunctionalInterface
      public interface Builder<V, B extends BossBar<V>> extends Facet<V> {
         @NotNull
         @NotNull B createBossBar(final @NotNull Collection<V> viewer);
      }
   }

   public interface Book<V, M, B> extends Message<V, M> {
      @Nullable B createBook(final @NotNull String title, final @NotNull String author, final @NotNull Iterable<M> pages);

      void openBook(final @NotNull V viewer, final @NotNull B book);
   }

   public interface EntitySound<V, M> extends Facet<V> {
      M createForSelf(final V viewer, final xyz.lychee.lagfixer.libs.adventure.sound.@NotNull Sound sound);

      M createForEmitter(final xyz.lychee.lagfixer.libs.adventure.sound.@NotNull Sound sound, final xyz.lychee.lagfixer.libs.adventure.sound.Sound.@NotNull Emitter emitter);

      void playSound(final @NotNull V viewer, final M message);
   }

   public interface Sound<V, P> extends Position<V, P> {
      void playSound(final @NotNull V viewer, final xyz.lychee.lagfixer.libs.adventure.sound.@NotNull Sound sound, final @NotNull P position);

      void stopSound(final @NotNull V viewer, final @NotNull SoundStop sound);
   }

   public interface Position<V, P> extends Facet<V> {
      @Nullable P createPosition(final @NotNull V viewer);

      @NotNull P createPosition(final double x, final double y, final double z);
   }

   public interface TitlePacket<V, M, C, T> extends Title<V, M, C, T> {
      int ACTION_TITLE = 0;
      int ACTION_SUBTITLE = 1;
      int ACTION_ACTIONBAR = 2;
      int ACTION_TIMES = 3;
      int ACTION_CLEAR = 4;
      int ACTION_RESET = 5;
   }

   public interface Title<V, M, C, T> extends Message<V, M> {
      int PROTOCOL_ACTION_BAR = 310;
      long MAX_SECONDS = 461168601842738790L;

      @NotNull C createTitleCollection();

      void contributeTitle(final @NotNull C coll, final @NotNull M title);

      void contributeSubtitle(final @NotNull C coll, final @NotNull M subtitle);

      void contributeTimes(final @NotNull C coll, final int inTicks, final int stayTicks, final int outTicks);

      @Nullable T completeTitle(final @NotNull C coll);

      void showTitle(final @NotNull V viewer, final @NotNull T title);

      void clearTitle(final @NotNull V viewer);

      void resetTitle(final @NotNull V viewer);

      default int toTicks(final @Nullable Duration duration) {
         if (duration != null && !duration.isNegative()) {
            return duration.getSeconds() > 461168601842738790L ? Integer.MAX_VALUE : (int)(duration.getSeconds() * 20L + (long)(duration.getNano() / 50000000));
         } else {
            return -1;
         }
      }
   }

   public interface ActionBar<V, M> extends Message<V, M> {
      void sendMessage(final @NotNull V viewer, final @NotNull M message);
   }

   public interface ChatPacket<V, M> extends Chat<V, M> {
      byte TYPE_CHAT = 0;
      byte TYPE_SYSTEM = 1;
      byte TYPE_ACTION_BAR = 2;

      default byte createMessageType(final @NotNull MessageType type) {
         if (type == MessageType.CHAT) {
            return 0;
         } else if (type == MessageType.SYSTEM) {
            return 1;
         } else {
            Knob.logUnsupported(this, type);
            return 0;
         }
      }
   }

   public interface Chat<V, M> extends Message<V, M> {
      void sendMessage(final @NotNull V viewer, final @NotNull Identity source, final @NotNull M message, final @NotNull Object type);
   }

   public interface Message<V, M> extends Facet<V> {
      int PROTOCOL_HEX_COLOR = 713;
      int PROTOCOL_JSON = 5;

      @Nullable M createMessage(final @NotNull V viewer, final @NotNull Component message);
   }
}
